/*
 *			P M O N . C
 *
 *  A program to moniter a process.
 *   18 July 85 - Phillip Dykstra
 *
 */
#include <curses.h>
#include <signal.h>
#include "pmon.h"

#ifndef	LPR
#define	LPR	"lpr"
#endif	LPR


#define	CNTRL(x)	((x) - 'A' + 1)

int done();

int monpid;
struct nlist namelist[] = {
	{ "_avenrun" },
	{ "_proc" },
	{ "_nproc" },
	{ "_Usrptmap" },
	{ "_usrpt" },
	{ 0 }
};

main(argc,argv)
int argc;
char *argv[];
{
        register i;
        double f1, f2;
        extern char *ctime();

	build_devtable();

	/*
	 * fill the namelist and open /dev/kmem
	 */
        kmem = open("/dev/kmem", 0);
	mem = open("/dev/mem", 0);
	swap = open("/dev/drum", 0);
	setgid(getgid());		/* for safety's sake */
	setuid(getuid());		/* in case it's installed this way */
        if (kmem < 0) {
                printf("cannot open /dev/kmem\n");
                exit(1);
        }
        if (mem < 0) {
                printf("cannot open /dev/mem\n");
                exit(1);
        }
        if (swap < 0) {
                printf("cannot open /dev/drum\n");
                exit(1);
        }
        nlist("/vmunix", namelist);

        if (namelist[X_NPROC].n_type == 0) {
                printf("nproc not found in /vmunix namelist!\n");
                exit(1);
        }

	/* if second argument take it to be a pid */
        if (argc > 1)
                monpid = (atoi(argv[1]));
        else
		monpid = getpid();

	/* if third argument take it to be an interval */
        if (argc > 2)
                intv = (atoi(argv[2]));
        else intv = 1;

	/* 
	 * do all things that need to be done only once
	 */
	usrpt = (struct pte *)namelist[X_USRPT].n_value;
	Usrptmap = (struct pte *)namelist[X_USRPTMAP].n_value;
	nproc = getw( namelist[X_NPROC].n_value );
	procp = (struct proc *)getw( namelist[X_PROC].n_value );

	if( findproc( monpid ) == 0 ) {
		fprintf( stderr, "pmon: No process with pid = %d\n", monpid );
		exit( 2 );
	}

	/*
	 * monitor process forever
	 */
        worker();
        /* NOTREACHED */
}

char obuf[BUFSIZ];	/* Output buffer */

worker()
{
        long	clock;
	int	i, tin;
	struct	timeval tintv;
	int	parent_pid = 1;
	int	in = 1 << fileno(stdin);
	int	number, building_number;

	/* set up signals */
	signal(SIGINT, done);
	signal(SIGQUIT, done);
/*signal(SIGTSTP, SIG_IGN);	/* Why doesn't this work? */
/*signal(SIGQUIT, SIG_IGN);	/* May wish to do this */

	/* set CBREAK mode with no buffering on stdin */
	setbuf(stdin, 0);
	setbuf(stdout, obuf);

	tintv.tv_sec = intv;
	tintv.tv_usec = 0;
	number = building_number = 0;
	initscr();	/* init curses package */
	crmode();	/* set terminal into CBREAK mode */
	noecho();
        dispinit();     /* initialize display */

        for( ;; ){
		/*
		 * Read in the proc structure.
        	 */
        	lseek( kmem, (long)procp, 0 );
		if( read(kmem, &pr, sizeof(pr)) != sizeof(pr) ) {
			printf("pmon: kernel read error!\n");
        		exit (0);
	        }
        	/*
		 * make sure it's valid.
        	 * If the process has exited we try to escape to parent.
        	 * If we can't find the parent, we go for init.
        	 */
#if	defined(ultrix) && !defined(SMASTER)
		if (allflags(&pr, 0) == 0) {
#else
        	if( pr.p_flag == 0 ) {
#endif	ultrix
        		printf("%c", CNTRL('G'));
        		if( findproc( parent_pid ) ) {
				mvprintw(LINES-1, 5, "Escaping to parent!\n");
			} else {
				findproc( 1 );
				mvprintw(LINES-1, 5, "Escaping to Init!\n");
        		}
        		refresh();
        		continue;
        	}
		parent_pid = pr.p_ppid;
        	if (getupage( &pr ) == 0) {
			dispinit();
			mvprintw(LINES-1, 5,
				"No such process, escaping to Init!\n");
			sleep(2);
			findproc( 1 );
			if (getupage(&pr) == 0) {
			     mvprintw(LINES-1, 5, "Giving up!\n");
			     exit(1);
			}
			dispinit();
		}

        	/* get and display load average */
		lseek(kmem, (long)namelist[LOADAV].n_value, 0);
#ifdef	_FIXPT_
		read(kmem, &fixloadavg[0], sizeof fixloadavg);
		for (i = 0; i < 2; i++)
		    loadavg[i] = FIX_TO_DBL(fixloadavg[i]);
#else
		read(kmem, &loadavg[0], sizeof loadavg);
#endif	_FIXPT_
#ifdef SUN
		mvprintw(0,13," %4.2f %4.2f %4.2f",
			loadavg[0]/256., loadavg[1]/256., loadavg[2]/256.);
#else SUN
#if	defined(vax) && !defined(ultrix)
		mvprintw(0,13," %4.2f %4.2f %4.2f %4.2f",
				loadavg[3], loadavg[0], loadavg[1], loadavg[2]);
#else
		mvprintw(0,13," %4.2f %4.2f %4.2f",
				loadavg[0], loadavg[1], loadavg[2]);
#endif vax
#endif SUN

        	/* Display Time */
		time(&clock);
		mvprintw(0,40,ctime(&clock));

		dispupdate();
		mvprintw(LINES-1, 0, "CMD> ");
		refresh();

		tin = 1;
		i = select(2, &tin, (int *)0, (int *)0, &tintv);
		if (i && tin) {
			i = getchar();
			if( i >= '0' && i <= '9' ) {
				if( building_number )
					number *= 10;
				number += i - '0';
				building_number++;
			}
			if( i == '\n' && building_number ) {
				if( findproc( number ) == 0 )
		        		printf("%c", CNTRL('G'));
				building_number = 0;
				number = 0;
			}
			switch(i)
				{
				FILE	*Outfile,*popen();

				case CNTRL('P'):	/* Parent */
					if( pr.p_pptr != 0 )
						procp = pr.p_pptr;
					break;
				case CNTRL('N'):	/* Child */
					if( pr.p_cptr != 0 )
						procp = pr.p_cptr;
					break;
				case CNTRL('F'):	/* Younger Sibling */
					if( pr.p_ysptr != 0 )
						procp = pr.p_ysptr;
					break;
				case CNTRL('B'):	/* Older Sibling */
					if( pr.p_osptr != 0 )
						procp = pr.p_osptr;
					break;
				case CNTRL('L'):	/* Repaint Screen */
					dispinit();
					break;
				case 'f':		/* Bump fd's */
					fdbase = (fdbase == 0) ? FDISPLAY : 0;
					break;
				case 'g':		/* Show proc groups */
					listgroup();
					break;
				case 't':		/* Show proc tree */
					clear();	/* clear screen */
					show_children(pr.p_cptr, 0);
					refresh();
					select( 1, &in, 0, 0, 0 );
					dispinit();
					break;
				case '?':		/* Display help */
					disphelp();
					break;
#ifdef	notdef
        			case 'l':
					/* Or "lpr" */
					Outfile = popen(LPR,"w");
					if (Outfile != 0 )
						{
/*						dumpit(Outfile);*/
						fclose(Outfile);
						}
					break;
        			case 's':
					Outfile = fopen("mon.dump","a");
					if (Outfile != 0 )
						{
/*						dumpit(Outfile);*/
						fclose(Outfile);
						}
					break;
#endif	notdef
        			case 'q':
        				done();
				}
        	}
        }
}

/*
 * DONE - put the term back in non CBREAK mode and exit.
 *   Note: on interrupts curses may be mixed up and thus the cursor
 *     can not be guaranteed to be left at the bottom of the screen.
 */
done()
{
	mvprintw(LINES-1,0,"\n");
	refresh();
	endwin();
	exit(0);
}

/*
 * G E T W - get a word from kernel memory
 */
getw(loc)
unsigned long loc;
{
	int word;

	lseek(kmem, (long)loc, 0);
	if (read(kmem, (char *)&word, sizeof(word)) != sizeof(word))
		printf("error reading kmem at %x\n", loc);

	return (word);
}
